require 'rexml/document'
require 'qmf'

module SchemaParser
  TYPES = {
    "bool" => Qmf::TYPE_BOOL,
    "double" => Qmf::TYPE_DOUBLE,
    "objId" => Qmf::TYPE_REF,
    "lstr" => Qmf::TYPE_LSTR,
    "sstr" => Qmf::TYPE_SSTR,
    "uint32" => Qmf::TYPE_UINT32,
    "uint64" => Qmf::TYPE_UINT64,
    "array" => Qmf::TYPE_ARRAY
  }

  ACCESS = {
    "RC" => Qmf::ACCESS_READ_CREATE,
    "RW" => Qmf::ACCESS_READ_WRITE,
    "RO" => Qmf::ACCESS_READ_ONLY,
    "R" => Qmf::ACCESS_READ_ONLY
  }

  DIRECTION = {
    "I" => Qmf::DIR_IN,
    "O" => Qmf::DIR_OUT,
    "IO" => Qmf::DIR_IN_OUT
  }

  def self.parse(args)
    doc = nil
    if args.has_key?(:xml)
      doc = REXML::Document.new(args[:xml])
    elsif args.has_key? :file
      File::open(args[:file], "r") { |fp| doc = REXML::Document.new(fp) }
    end

    package = doc.root.attributes["package"]
    #puts "Package: #{package}"
    klasses = []
    REXML::XPath.each(doc.root, "/schema/class") do |ele|
      klasses << parse_class(package, ele)
    end
    return klasses
  end

  private

  def self.parse_class(package, ele)
    name = ele.attributes["name"]
    qmf_class = Qmf::SchemaObjectClass.new(package, name)
    #puts "Class: #{name}"
    REXML::XPath.each(ele, "property") do |p|
      qmf_class.add_property(parse_property(p))
    end
    REXML::XPath.each(ele, "method") do |m|
      qmf_class.add_method(parse_method(m))
    end
    return qmf_class
  end

  def self.parse_property(ele)
    hash = {}
    name = ele.attributes["name"]
    type = typecode(ele.attributes["type"])
    hash = attribute_hash(ele, [:access, :index, :optional, :unit, :desc])
    #puts "  Property: #{name} : #{type.inspect} #{hash.inspect}"
    Qmf::SchemaProperty.new(name, type, hash)
  end

  def self.parse_method(ele)
    name = ele.attributes["name"]
    hash = attribute_hash(ele, [:desc])
    #puts "  Method: #{name} #{hash.inspect}"
    method = Qmf::SchemaMethod.new(name, hash)
    REXML::XPath.each(ele, "arg") do |a|
      method.add_argument(parse_arg(a))
    end
    method
  end

  def self.parse_arg(ele)
    name = ele.attributes["name"]
    type = typecode(ele.attributes["type"])
    hash = attribute_hash(ele, [:dir, :unit, :desc])
    #puts "    #{name} : #{type} #{hash.inspect}"
    Qmf::SchemaArgument.new(name, type, hash)
  end

  def self.attribute_hash(ele, attrs)
    attrs.inject({}) do |hash, kw|
      if v = ele.attributes[kw.to_s]
        # FIXME: Does :unit need conversion ?
        v = index(v) if kw == :index
        v = access(v) if kw == :access
        v = direction(v) if kw == :dir
        hash[kw] = v
      end
      hash
    end
  end

  def self.typecode(typename)
    enum_val(TYPES, typename, "type")
  end

  def self.index(ind)
    # convert to bool
    return true if ind == true || ind =~ /^true$/i || ind =~ /^y$/i
    return false if ind == false || ind.nil? || ind =~ /^false$/i || ind =~ /^n$/i
    raise ArgumentError, "invalid value for Boolean: \"#{self}\""
  end

  def self.access(acc)
    enum_val(ACCESS, acc, "access code")
  end

  def self.direction(dir)
    enum_val(DIRECTION, dir, "direction")
  end

  def self.enum_val(enum, name, desc)
    unless val = enum[name]
      raise "Unknown #{desc} #{name}"
    end
    return val
  end
end
