
#include <string.h>
#include <sys/select.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <syslog.h>
#include <signal.h>

#include "StressAgent.h"

namespace _qmf = qmf::com::redhat::stress_agent;

StressAgent::StressAgent(ManagementAgent* _agent, string _name) : name(_name), agent(_agent)
{
    char hostname[1024];

    gethostname(hostname, sizeof(hostname));

    mgmtObject = new _qmf::Stress(agent, this, hostname);
    agent->addObject(mgmtObject);
}

StressAgent::~StressAgent()
{
    mgmtObject->resourceDestroy();
}

void StressAgent::create_test_objects(int test_count)
{
    int i;

    for(i = 0; i < test_count; i++) {
        Test *test_obj = new Test(agent, this, i);
        test_objects.push_back(test_obj);
    }
}


void StressAgent::do_loop()
{
    fd_set fds;
    struct timeval tv;
    int retval;

    /* Go through all domains and call update() for each, letting them update
     * information and statistics. */
    while (1) {
        int read_fd = agent->getSignalFd();

        for (std::vector<Test*>::iterator iter = test_objects.begin();
                iter != test_objects.end(); iter++) {
            (*iter)->update();
        }

        /* Poll agent fd.  If any methods are being made this FD will be ready for reading.  */
        FD_ZERO(&fds);
        FD_SET(read_fd, &fds);

        /* Wait up to three seconds. */
        tv.tv_sec = 3;
        tv.tv_usec = 0;

        retval = select(read_fd + 1, &fds, NULL, NULL, &tv);
        if (retval < 0) {
            fprintf(stderr, "Error in select loop: %s\n", strerror(errno));
            continue;
        }

        if (retval > 0) {
            /* This implements any pending methods. */
            agent->pollCallbacks();
        }
    }
}

Manageable::status_t
StressAgent::ManagementMethod(uint32_t methodId, Args& args, std::string &errstr)
{
    cout << "Method Received: " << methodId << endl;
    int ret;

    /*
    switch (methodId) {
    }
    */

    return STATUS_NOT_IMPLEMENTED;
}

static void
print_usage()
{
    printf("Usage:\tlibvirt-qpid <options>\n");
    printf("\t-d | --daemon     run as a daemon.\n");
    printf("\t-h | --help       print this help message.\n");
    printf("\t-b | --broker     specify broker host name..\n");
    printf("\t-g | --gssapi     force GSSAPI authentication.\n");
    printf("\t-u | --username   username to use for authentication purproses.\n");
    printf("\t-s | --service    service name to use for authentication purproses.\n");
    printf("\t-p | --port       specify broker port.\n");
    printf("\t-n | --num        specify number of objects to create.\n");
}


//==============================================================
// Main program
//==============================================================
int main(int argc, char** argv) {
    int arg;
    int idx = 0;
    bool verbose = true;
    bool daemonize = false;
    bool gssapi = false;
    char *host = NULL;
    char *username = NULL;
    char *service = NULL;
    int port = 5672;
    int num_obj = 500;

    struct option opt[] = {
        {"help", 0, 0, 'h'},
        {"daemon", 0, 0, 'd'},
        {"broker", 1, 0, 'b'},
        {"gssapi", 0, 0, 'g'},
        {"username", 1, 0, 'u'},
        {"service", 1, 0, 's'},
        {"port", 1, 0, 'p'},
        {"num", 1, 0, 'n'},
        {0, 0, 0, 0}
    };

    while ((arg = getopt_long(argc, argv, "hdb:gu:s:n:p:", opt, &idx)) != -1) {
        switch (arg) {
            case 'h':
            case '?':
                print_usage();
                exit(0);
                break;
            case 'd':
                daemonize = true;
                break;
            case 'v':
                verbose = true;
                break;
            case 's':
                if (optarg) {
                    service = strdup(optarg);
                } else {
                    print_usage();
                    exit(1);
                }
                break;
            case 'u':
                if (optarg) {
                    username = strdup(optarg);
                } else {
                    print_usage();
                    exit(1);
                }
                break;
            case 'g':
                gssapi = true;
                break;
            case 'p':
                if (optarg) {
                    port = atoi(optarg);
                } else {
                    print_usage();
                    exit(1);
                }
                break;
            case 'n':
                if (optarg) {
                    num_obj = atoi(optarg);
                } else {
                    print_usage();
                    exit(1);
                }
                break;
            case 'b':
                if (optarg) {
                    host = strdup(optarg);
                } else {
                    print_usage();
                    exit(1);
                }
                break;
            default:
                fprintf(stderr, "unsupported option '-%c'.  See --help.\n", arg);
                print_usage();
                exit(0);
            break;
        }
    }

    if (daemonize == true) {
        if (daemon(0, 0) < 0) {
            fprintf(stderr, "Error daemonizing: %s\n", strerror(errno));
            exit(1);
        }
    }
    openlog("stress_agent", 0, LOG_DAEMON);

    // This prevents us from dying if libvirt disconnects.
    signal(SIGPIPE, SIG_IGN);

    // Create the management agent
    ManagementAgent::Singleton singleton;
    ManagementAgent* agent = singleton.getInstance();

    // Register the schema with the agent
    _qmf::Package packageInit(agent);

    // Start the agent.  It will attempt to make a connection to the
    // management broker.  The third argument is the interval for sending
    // updates to stats/properties to the broker.  The last is set to 'true'
    // to keep this all single threaded.  Otherwise a second thread would be
    // used to implement methods.

    qpid::client::ConnectionSettings settings;
    settings.host = host ? host : "127.0.0.1";
    settings.port = port;

    if (username != NULL) {
        settings.username = username;
    }
    if (service != NULL) {
        settings.service = service;
    }
    if (gssapi == true) {
        settings.mechanism = "GSSAPI";
    }

    agent->init(settings.host, settings.port, 3, true, "", settings.username, "", settings.mechanism);

    while(true) {
        try {
            StressAgent stresser(agent, "QMF Agent Stress Tester");
            stresser.create_test_objects(num_obj);
            stresser.do_loop();
        } catch (int err) {
        }
        sleep(10);
    }
}


